﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Linq;
using System.Threading;
using System.Timers;
using HeuristicLab.Collections;
using HeuristicLab.Common;
using HeuristicLab.Common.Resources;
using HeuristicLab.Core;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using Timer = System.Timers.Timer;

namespace HeuristicLab.Optimization {
  internal enum TimeLimitedRunAction { None, Prepare, Start, Pause, Stop };

  /// <summary>
  /// A run in which an algorithm is executed for a certain maximum time only.
  /// </summary>
  [Item("TimeLimitRun", "A run in which an optimizer is executed a certain maximum time.")]
  [Creatable("Testing & Analysis")]
  [StorableClass]
  public sealed class TimeLimitRun : NamedItem, IOptimizer, IStorableContent, INotifyPropertyChanged {
    public string Filename { get; set; }

    #region ItemImage
    public static new Image StaticItemImage {
      get { return VSImageLibrary.Event; }
    }
    public override Image ItemImage {
      get { return (Algorithm != null) ? Algorithm.ItemImage : VSImageLibrary.ExecutableStopped; }
    }
    #endregion

    private readonly Timer snapTimer;
    private readonly Timer runTimer;
    private readonly ManualResetEvent algorithmPaused = new ManualResetEvent(false);
    private readonly ManualResetEvent algorithmStopped = new ManualResetEvent(false);

    [Storable]
    private TimeSpan maximumExecutionTime;
    public TimeSpan MaximumExecutionTime {
      get { return maximumExecutionTime; }
      set {
        if (maximumExecutionTime == value) return;
        maximumExecutionTime = value;
        runTimer.Interval = (maximumExecutionTime - ExecutionTime).TotalMilliseconds;
        OnPropertyChanged("MaximumExecutionTime");
      }
    }

    [Storable]
    private int snapshotTimesIndex;
    [Storable]
    private ObservableList<double> snapshotTimes;
    public ObservableList<double> SnapshotTimes {
      get { return snapshotTimes; }
      set {
        if (snapshotTimes == value) return;
        snapshotTimes = value;
        snapshotTimesIndex = 0;
        OnPropertyChanged("SnapshotTimes");
      }
    }

    [Storable]
    private bool storeAlgorithmInEachSnapshot;
    [Storable]
    public bool StoreAlgorithmInEachSnapshot {
      get { return storeAlgorithmInEachSnapshot; }
      set {
        if (storeAlgorithmInEachSnapshot == value) return;
        storeAlgorithmInEachSnapshot = value;
        OnPropertyChanged("StoreAlgorithmInEachSnapshot");
      }
    }

    [Storable]
    private RunCollection snapshots;
    public RunCollection Snapshots {
      get { return snapshots; }
      set {
        if (snapshots == value) return;
        snapshots = value;
        OnPropertyChanged("Snapshots");
      }
    }

    #region Inherited Properties
    public ExecutionState ExecutionState {
      get { return (Algorithm != null) ? Algorithm.ExecutionState : ExecutionState.Stopped; }
    }

    public TimeSpan ExecutionTime {
      get { return (Algorithm != null) ? Algorithm.ExecutionTime : TimeSpan.FromSeconds(0); }
    }

    [Storable]
    private IAlgorithm algorithm;
    public IAlgorithm Algorithm {
      get { return algorithm; }
      set {
        if (algorithm == value) return;
        if (algorithm != null) {
          DeregisterAlgorithmEvents();
          var runs = algorithm.Runs;
          algorithm = null; //necessary to avoid removing the runs from the old algorithm
          Runs.RemoveRange(runs);
        }
        algorithm = value;
        if (algorithm != null) {
          RegisterAlgorithmEvents();
          Runs.AddRange(algorithm.Runs);
          if (algorithm.ExecutionState == ExecutionState.Paused) algorithmPaused.Set();
          else algorithmPaused.Reset();
          if (algorithm.ExecutionState == ExecutionState.Stopped) algorithmStopped.Set();
          else algorithmStopped.Reset();
        }
        OnPropertyChanged("Algorithm");
        Prepare();
      }
    }

    [Storable]
    private RunCollection runs;
    public RunCollection Runs {
      get { return runs; }
      private set {
        if (value == null) throw new ArgumentNullException();
        if (runs == value) return;
        runs = value;
        OnPropertyChanged("Runs");
      }
    }

    public IEnumerable<IOptimizer> NestedOptimizers {
      get {
        if (Algorithm == null) yield break;
        yield return Algorithm;
        foreach (var opt in Algorithm.NestedOptimizers)
          yield return opt;
      }
    }
    #endregion

    [StorableConstructor]
    private TimeLimitRun(bool deserializing)
      : base(deserializing) {
      // the timer will not be serialized
      snapTimer = new Timer() {
        AutoReset = false,
        Enabled = false,
        Interval = double.MaxValue
      };
      runTimer = new Timer() {
        AutoReset = false,
        Enabled = false,
        Interval = double.MaxValue
      };
    }
    private TimeLimitRun(TimeLimitRun original, Cloner cloner)
      : base(original, cloner) {
      maximumExecutionTime = original.maximumExecutionTime;
      snapshotTimes = new ObservableList<double>(original.snapshotTimes);
      snapshotTimesIndex = original.snapshotTimesIndex;
      snapTimer = new Timer() {
        AutoReset = original.snapTimer.AutoReset,
        Enabled = original.snapTimer.Enabled,
        Interval = original.snapTimer.Interval
      };
      runTimer = new Timer() {
        AutoReset = original.runTimer.AutoReset,
        Enabled = original.runTimer.Enabled,
        Interval = original.runTimer.Interval
      };
      snapshots = cloner.Clone(original.snapshots);
      storeAlgorithmInEachSnapshot = original.storeAlgorithmInEachSnapshot;
      algorithm = cloner.Clone(original.algorithm);
      runs = cloner.Clone(original.runs);
      Initialize();
    }
    public TimeLimitRun()
      : base() {
      name = ItemName;
      description = ItemDescription;
      maximumExecutionTime = TimeSpan.FromMinutes(1);
      snapshotTimes = new ObservableList<double>(new[] {
          TimeSpan.FromSeconds(5).TotalMilliseconds,
          TimeSpan.FromSeconds(10).TotalMilliseconds,
          TimeSpan.FromSeconds(30).TotalMilliseconds });
      snapshotTimesIndex = 0;
      snapTimer = new Timer() {
        AutoReset = false,
        Enabled = false,
        Interval = snapshotTimes[snapshotTimesIndex]
      };
      runTimer = new Timer() {
        AutoReset = false,
        Enabled = false,
        Interval = MaximumExecutionTime.TotalMilliseconds
      };
      snapshots = new RunCollection();
      Runs = new RunCollection { AlgorithmName = Name };
    }
    public TimeLimitRun(string name)
      : base(name) {
      description = ItemDescription;
      maximumExecutionTime = TimeSpan.FromMinutes(1);
      snapshotTimes = new ObservableList<double>(new[] {
          TimeSpan.FromSeconds(5).TotalMilliseconds,
          TimeSpan.FromSeconds(10).TotalMilliseconds,
          TimeSpan.FromSeconds(30).TotalMilliseconds });
      snapshotTimesIndex = 0;
      snapTimer = new Timer() {
        AutoReset = false,
        Enabled = false,
        Interval = snapshotTimes[snapshotTimesIndex]
      };
      snapshots = new RunCollection();
      runTimer = new Timer() {
        AutoReset = false,
        Enabled = false,
        Interval = MaximumExecutionTime.TotalMilliseconds
      };
      Runs = new RunCollection { AlgorithmName = Name };
    }
    public TimeLimitRun(string name, string description)
      : base(name, description) {
      maximumExecutionTime = TimeSpan.FromMinutes(1);
      snapshotTimes = new ObservableList<double>(new[] {
          TimeSpan.FromSeconds(5).TotalMilliseconds,
          TimeSpan.FromSeconds(10).TotalMilliseconds,
          TimeSpan.FromSeconds(30).TotalMilliseconds });
      snapshotTimesIndex = 0;
      snapTimer = new Timer() {
        AutoReset = false,
        Enabled = false,
        Interval = snapshotTimes[snapshotTimesIndex]
      };
      snapshots = new RunCollection();
      runTimer = new Timer() {
        AutoReset = false,
        Enabled = false,
        Interval = MaximumExecutionTime.TotalMilliseconds
      };
      Runs = new RunCollection { AlgorithmName = Name };
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      if (ExecutionState == ExecutionState.Started) throw new InvalidOperationException(string.Format("Clone not allowed in execution state \"{0}\".", ExecutionState));
      return new TimeLimitRun(this, cloner);
    }

    [StorableHook(HookType.AfterDeserialization)]
    private void AfterDeserialization() {
      Initialize();
    }

    private void Initialize() {
      if (algorithm != null) RegisterAlgorithmEvents();
      snapTimer.Elapsed += snapshotTimer_Elapsed;
      runTimer.Elapsed += runTimer_Elapsed;
    }

    private readonly object timerLock = new object();
    private void snapshotTimer_Elapsed(object sender, ElapsedEventArgs e) {
      lock (timerLock) {
        runTimer.Stop();
        if (Algorithm == null) return;
        if (ExecutionState != ExecutionState.Started) return;
        var error = false;
        try { Algorithm.Pause(); } catch (InvalidOperationException) { error = true; }
        if (error && ExecutionState != ExecutionState.Paused) return; // Algorithm is either stopped or prepared

        algorithmPaused.WaitOne();
        MakeSnapshot();
        try { Algorithm.Start(); } catch (InvalidOperationException) { }
      }
    }

    private void runTimer_Elapsed(object sender, ElapsedEventArgs e) {
      lock (timerLock) {
        runTimer.Stop();
        snapTimer.Stop();
        try { Algorithm.Stop(); } catch (InvalidOperationException) { }
      }
    }

    private double? GetNextInterval(TimeSpan reference) {
      double interval;
      while ((interval = snapshotTimes[snapshotTimesIndex] - reference.TotalMilliseconds) <= 0) {
        snapshotTimesIndex++;
        if (snapshotTimesIndex >= snapshotTimes.Count) return null;
      };
      return interval;
    }

    private void MakeSnapshot() {
      string runName = ExecutionTime.ToString() + " " + algorithm.Name + " Snapshot";
      Run run;
      if (StoreAlgorithmInEachSnapshot) {
        run = new Run(runName, algorithm);
      } else {
        // duplicate code of Run.Initialize - necessary, because Algorithm property doesn't have a setter
        run = new Run() { Name = runName };
        var parameters = new Dictionary<string, IItem>();
        var results = new Dictionary<string, IItem>();
        algorithm.CollectParameterValues(parameters);
        algorithm.CollectResultValues(results);
        var cloner = new Cloner();
        foreach (var kvp in parameters.Select(x => new KeyValuePair<string, IItem>(x.Key, cloner.Clone(x.Value))).ToDictionary(x => x.Key, x => x.Value))
          run.Parameters.Add(kvp);
        foreach (var kvp in results.Select(x => new KeyValuePair<string, IItem>(x.Key, cloner.Clone(x.Value))).ToDictionary(x => x.Key, x => x.Value))
          run.Results.Add(kvp);
      }
      snapshots.Add(run);
    }

    public void Prepare() {
      Prepare(false);
    }
    public void Prepare(bool clearRuns) {
      Algorithm.Prepare(clearRuns);
    }
    public void Start() {
      Algorithm.Start();
    }
    public void Pause() {
      Algorithm.Pause();
    }
    public void Stop() {
      Algorithm.Stop();
    }

    #region Events
    protected override void OnNameChanged() {
      base.OnNameChanged();
      runs.AlgorithmName = Name;
    }

    public event PropertyChangedEventHandler PropertyChanged;
    private void OnPropertyChanged(string property) {
      var handler = PropertyChanged;
      if (handler != null) handler(this, new PropertyChangedEventArgs(property));
    }

    #region IExecutable Events
    public event EventHandler ExecutionStateChanged;
    private void OnExecutionStateChanged() {
      var handler = ExecutionStateChanged;
      if (handler != null) handler(this, EventArgs.Empty);
    }
    public event EventHandler ExecutionTimeChanged;
    private void OnExecutionTimeChanged() {
      var handler = ExecutionTimeChanged;
      if (handler != null) handler(this, EventArgs.Empty);
    }
    public event EventHandler Prepared;
    private void OnPrepared() {
      var handler = Prepared;
      if (handler != null) handler(this, EventArgs.Empty);
    }
    public event EventHandler Started;
    private void OnStarted() {
      var handler = Started;
      if (handler != null) handler(this, EventArgs.Empty);
    }
    public event EventHandler Paused;
    private void OnPaused() {
      var handler = Paused;
      if (handler != null) handler(this, EventArgs.Empty);
    }
    public event EventHandler Stopped;
    private void OnStopped() {
      var handler = Stopped;
      if (handler != null) handler(this, EventArgs.Empty);
    }
    public event EventHandler<EventArgs<Exception>> ExceptionOccurred;
    private void OnExceptionOccurred(Exception exception) {
      var handler = ExceptionOccurred;
      if (handler != null) handler(this, new EventArgs<Exception>(exception));
    }
    #endregion

    #region Algorithm Events
    private void RegisterAlgorithmEvents() {
      algorithm.ExceptionOccurred += Algorithm_ExceptionOccurred;
      algorithm.ExecutionTimeChanged += Algorithm_ExecutionTimeChanged;
      algorithm.ExecutionStateChanged += Algorithm_ExecutionStateChanged;
      algorithm.Paused += Algorithm_Paused;
      algorithm.Prepared += Algorithm_Prepared;
      algorithm.Started += Algorithm_Started;
      algorithm.Stopped += Algorithm_Stopped;
    }
    private void DeregisterAlgorithmEvents() {
      algorithm.ExceptionOccurred -= Algorithm_ExceptionOccurred;
      algorithm.ExecutionTimeChanged -= Algorithm_ExecutionTimeChanged;
      algorithm.ExecutionStateChanged -= Algorithm_ExecutionStateChanged;
      algorithm.Paused -= Algorithm_Paused;
      algorithm.Prepared -= Algorithm_Prepared;
      algorithm.Started -= Algorithm_Started;
      algorithm.Stopped -= Algorithm_Stopped;
    }
    private void Algorithm_ExceptionOccurred(object sender, EventArgs<Exception> e) {
      OnExceptionOccurred(e.Value);
    }
    private void Algorithm_ExecutionTimeChanged(object sender, EventArgs e) {
      OnExecutionTimeChanged();
    }
    private void Algorithm_ExecutionStateChanged(object sender, EventArgs e) {
      OnExecutionStateChanged();
    }
    private void Algorithm_Paused(object sender, EventArgs e) {
      algorithmPaused.Set();
      algorithmStopped.Reset();
      OnPaused();
    }
    private void Algorithm_Prepared(object sender, EventArgs e) {
      algorithmPaused.Reset();
      algorithmStopped.Reset();
      snapTimer.Stop();
      runTimer.Stop();
      snapshotTimesIndex = 0;
      snapshots.Clear();
      OnPrepared();
    }
    private void Algorithm_Started(object sender, EventArgs e) {
      algorithmPaused.Reset();
      algorithmStopped.Reset();
      var execTime = ExecutionTime;
      var interval = GetNextInterval(execTime);
      if (interval.HasValue) {
        snapTimer.Interval = interval.Value;
        snapTimer.Start();
      }
      runTimer.Interval = (MaximumExecutionTime - execTime).TotalMilliseconds;
      runTimer.Start();
      OnStarted();
    }
    private void Algorithm_Stopped(object sender, EventArgs e) {
      algorithmPaused.Reset();
      algorithmStopped.Set();
      snapTimer.Stop();
      runTimer.Stop();
      MakeSnapshot();
      var snapshotsRun = new Run() { Name = algorithm.Runs.Last().Name };
      foreach (var s in snapshots)
        snapshotsRun.Parameters.Add(s.Name, s);
      Runs.Add(snapshotsRun);
      OnStopped();
    }
    #endregion
    #endregion
  }
}
